//
// Copyright 2024 Staysail Systems, Inc. <info@staysail.tech>
//
// This software is supplied under the terms of the MIT License, a
// copy of which should be located in the distribution where this
// file was obtained (LICENSE.txt).  A copy of the license may also be
// found online at https://opensource.org/licenses/MIT.
//

#ifndef NNG_SUPPLEMENTAL_UTIL_PLATFORM_H
#define NNG_SUPPLEMENTAL_UTIL_PLATFORM_H

// The declarations in this file have been relocated to <nng/nng.h>.
// This file itself is deprecated, and applications should stop including
// it directly.
#include <nng/nng.h>

#endif // NNG_SUPPLEMENTAL_UTIL_PLATFORM_H
